# This file is part of sphinx-terminal.
#
# Copyright 2025 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 3, as published by the Free Software
# Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranties of MERCHANTABILITY, SATISFACTORY
# QUALITY, or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

"""Adds the directive to Sphinx."""

import importlib.util

from sphinx.util.typing import ExtensionMetadata
from sphinx.application import Sphinx
from .directive import TerminalDirective
from sphinx_terminal import common

try:
    from ._version import __version__
except ImportError:
    from importlib.metadata import version, PackageNotFoundError

    try:
        __version__ = version("sphinx_terminal")
    except PackageNotFoundError:
        __version__ = "dev"


def setup(app: Sphinx) -> ExtensionMetadata:
    """Connect the extension to the Sphinx application instance.

    app (Sphinx): The Sphinx application instance.

    returns: ExtensionMetadata
    """
    # add sphinx-copybutton to Sphinx's extensions
    try:
        if importlib.util.find_spec("sphinx_copybutton") is not None:
            app.setup_extension("sphinx_copybutton")
    except ModuleNotFoundError:
        print("Could not find 'sphinx-copybutton'.")

    app.add_directive("terminal", TerminalDirective)

    common.add_css(app, "terminal.css")

    copybutton_classes = "span.copybutton"
    if "copybutton_selector" not in app.config.values:
        app.add_config_value("copybutton_selector", copybutton_classes, "html")
    if app.config.copybutton_selector == "div.highlight pre":
        app.config.copybutton_selector = copybutton_classes

    # Configure copybutton to ignore the multiline prompt. These values will be
    # overridden if included in conf.py
    if "copybutton_prompt_text" not in app.config.values:
        app.add_config_value("copybutton_prompt_text", default="> |", rebuild="html")
    if "copybutton_prompt_is_regexp" not in app.config.values:
        app.add_config_value(
            "copybutton_prompt_is_regexp", default=True, rebuild="html"
        )

    return {
        "version": __version__,
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }


__all__ = ["__version__", "setup"]



import pytest
from sphinx_terminal import TerminalDirective
from typing_extensions import override


class FakeTerminalDirective(TerminalDirective):
    @override
    def __init__(self, options, content):
        self.options = options
        self.content = content


@pytest.fixture
def fake_terminal_directive(request: pytest.FixtureRequest) -> FakeTerminalDirective:
    """This fixture can be parametrized to override the default values."""
    # Get any optional overrides from the fixtures
    overrides = request.param if hasattr(request, "param") else {}

    return FakeTerminalDirective(
        options=overrides.get("options", {}),
        content=overrides.get("content"),
    )



# This file is part of sphinx-terminal.
#
# Copyright 2025 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 3, as published by the Free Software
# Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranties of MERCHANTABILITY, SATISFACTORY
# QUALITY, or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

"""Define common function for copying static asset files."""

from pathlib import Path

from sphinx.application import Sphinx
from sphinx.util.fileutil import copy_asset_file


def copy_custom_files(app: Sphinx, exc: Exception | None, filename: str) -> None:
    """Copy the specified file from the _static directory to the build directory.

    app (Sphinx): Sphinx application instance

    exc (Exception | None):

    filename (str):
    """
    if not exc and app.builder.format == "html":
        cssfile = Path(__file__).parent / "_static" / filename
        staticfile = app.builder.outdir / "_static" / filename
        copy_asset_file(cssfile, staticfile)


def add_css(app: Sphinx, filename: str) -> None:
    """Add a CSS file to the static directory.

    app (Sphinx):

    filename (str) :
    """

    def shim(app: Sphinx, exc: Exception | None) -> None:
        return copy_custom_files(app, exc, filename)

    app.connect("build-finished", shim)  # type: ignore[reportUnknownMemberType]
    app.add_css_file(filename)


def add_js(app: Sphinx, filename: str) -> None:
    """Add a javascript file to the static directory.

    app (Sphinx):

    filename (str) :
    """

    def shim(app: Sphinx, exc: Exception | None) -> None:
        return copy_custom_files(app, exc, filename)

    app.connect("build-finished", shim)  # type: ignore[reportUnknownMemberType]
    app.add_js_file(filename)



# This file is part of sphinx-terminal.
#
# Copyright 2025 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 3, as published by the Free Software
# Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranties of MERCHANTABILITY, SATISFACTORY
# QUALITY, or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import pathlib
import sys

project_dir = pathlib.Path(__file__).parents[1].resolve()
sys.path.insert(0, str(project_dir.absolute()))

project = "Terminal test output"
author = "Tester"

html_title = project
html_theme = "furo"

extensions = [
    "sphinx_terminal",
    "myst_parser",
]

copybutton_prompt_text = "> |"
copybutton_prompt_is_regexp = True



# This file is part of sphinx-terminal.
#
# Copyright 2025 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 3, as published by the Free Software
# Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranties of MERCHANTABILITY, SATISFACTORY
# QUALITY, or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import shutil
import subprocess
from pathlib import Path
from typing import cast

import bs4
import pytest


@pytest.fixture
def example_project(request) -> Path:
    project_root = request.config.rootpath
    example_dir = project_root / "tests/integration/example"

    target_dir = Path().resolve() / "example"
    shutil.copytree(example_dir, target_dir, dirs_exist_ok=True)

    return target_dir


@pytest.mark.slow
def test_hello_integration(example_project):
    build_dir = example_project / "_build"
    subprocess.check_call(
        ["sphinx-build", "-b", "html", "-W", example_project, build_dir],
    )

    index = build_dir / "index.html"

    # Rename the test output to something more meaningful
    shutil.copytree(
        build_dir, build_dir.parents[1] / ".test_output", dirs_exist_ok=True
    )
    soup = bs4.BeautifulSoup(index.read_text(), features="lxml")

    shutil.rmtree(example_project)  # Delete copied source

    # Ensure that the :copy: and :scroll: options are respected
    assert soup.find("div", {"class": "terminal scroll docutils container"})

    # Ensure that the prompt renders correctly
    prompt_html = soup.find("span", {"class": "pre"})
    assert getattr(prompt_html, "text", "") == "author@canonical:~/path$"

    # Ensure that the input command renders correctly
    input_html = soup.find("span", {"class": "command copybutton"})
    if input_html:
        command = cast(bs4.Tag, input_html).find_all("span", {"class": "pre"})
        assert getattr(command[0], "text", "") == "input"
    else:
        pytest.fail("Input command isn't present in rendered output.")

    output_html = soup.find(
        "div", {"class": "terminal-code highlight-text notranslate"}
    )
    if output_html:
        output = cast(bs4.Tag, output_html).find_next("pre")
        assert getattr(output, "text", "") == "output\n"
    else:
        pytest.fail("Command output is not rendered in output.")

    # Ensure that the prompt isn't inserted without an initial input line
    input_html = soup.find_next(
        "code", {"class": "command docutils literal notranslate"}
    )
    assert not input_html



# This file is part of sphinx-terminal.
#
# Copyright 2025 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 3, as published by the Free Software
# Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranties of MERCHANTABILITY, SATISFACTORY
# QUALITY, or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

import pytest
from docutils import nodes
from sphinx import addnodes
from sphinx.errors import SphinxError


@pytest.mark.parametrize(
    "fake_terminal_directive",
    [{"options": {}, "content": ["echo 'hello'"]}],
    indirect=True,
)
def test_terminal_directive(fake_terminal_directive):
    expected = nodes.container()
    expected["classes"] = "terminal"

    highlight = addnodes.highlightlang()
    highlight["force"] = "False"
    highlight["lang"] = "text"
    highlight["linenothreshold"] = "10000"
    expected.append(highlight)

    input_container = nodes.container()
    input_container["classes"] = "input"

    prompt_container = nodes.container()
    prompt_container["classes"] = "prompt"
    prompt_text = nodes.literal(text="user@host:~$ ")
    prompt_container.append(prompt_text)
    input_container.append(prompt_container)

    command_container = nodes.inline()
    command_container["classes"] = "command"
    command = nodes.literal(text="echo 'hello'\n")
    command_container.append(command)
    input_container.append(command_container)
    expected.append(input_container)

    actual = fake_terminal_directive.run()[0]

    assert str(expected) == str(actual)


@pytest.mark.parametrize(
    "fake_terminal_directive",
    [
        {
            "options": {
                "user": "author",
                "host": "canonical",
                "dir": "~/path",
            },
            "content": ["echo 'hello'", "", "hello"],
        }
    ],
    indirect=True,
)
def test_terminal_directive_prompt(fake_terminal_directive):
    expected = nodes.container()
    expected["classes"] = "terminal"

    highlight = addnodes.highlightlang()
    highlight["force"] = "False"
    highlight["lang"] = "text"
    highlight["linenothreshold"] = "10000"
    expected.append(highlight)

    input_container = nodes.container()
    input_container["classes"] = "input"

    prompt_container = nodes.container()
    prompt_container["classes"] = "prompt"
    prompt_text = nodes.literal(text="author@canonical:~/path$ ")
    prompt_container.append(prompt_text)
    input_container.append(prompt_container)

    command_container = nodes.inline()
    command_container["classes"] = "command"
    command = nodes.literal(text="echo 'hello'\n")
    command_container.append(command)
    input_container.append(command_container)
    expected.append(input_container)

    output_block = nodes.literal_block(text="hello")
    output_block["classes"] = "terminal-code"
    output_block["xml:space"] = "preserve"
    expected.append(output_block)

    actual = fake_terminal_directive.run()[0]

    assert str(expected) == str(actual)


@pytest.mark.parametrize(
    "fake_terminal_directive",
    [
        {
            "options": {
                "copy": None,
                "scroll": None,
            },
            "content": ["echo 'hello'", "", "hello"],
        }
    ],
    indirect=True,
)
def test_terminal_copy_scroll(fake_terminal_directive):
    expected = nodes.container()
    expected["classes"] = "terminal scroll"

    highlight = addnodes.highlightlang()
    highlight["force"] = "False"
    highlight["lang"] = "text"
    highlight["linenothreshold"] = "10000"
    expected.append(highlight)

    input_container = nodes.container()
    input_container["classes"] = "input"

    prompt_container = nodes.container()
    prompt_container["classes"] = "prompt"
    prompt_text = nodes.literal(text="user@host:~$ ")
    prompt_container.append(prompt_text)
    input_container.append(prompt_container)

    command_container = nodes.inline()
    command_container["classes"] = "command copybutton"
    command = nodes.literal(text="echo 'hello'\n")
    command_container.append(command)
    input_container.append(command_container)
    expected.append(input_container)

    output_block = nodes.literal_block(text="hello")
    output_block["classes"] = "terminal-code"
    output_block["xml:space"] = "preserve"
    expected.append(output_block)

    actual = fake_terminal_directive.run()[0]

    assert str(expected) == str(actual)


@pytest.mark.parametrize(
    "fake_terminal_directive",
    [
        {
            "options": {"class": ["test"]},
            "content": ["echo 'hello'", "", "hello"],
        }
    ],
    indirect=True,
)
def test_terminal_class_option(fake_terminal_directive):
    expected = nodes.container()
    expected["classes"] = "terminal test"

    highlight = addnodes.highlightlang()
    highlight["force"] = "False"
    highlight["lang"] = "text"
    highlight["linenothreshold"] = "10000"
    expected.append(highlight)

    input_container = nodes.container()
    input_container["classes"] = "input"

    prompt_container = nodes.container()
    prompt_container["classes"] = "prompt"
    prompt_text = nodes.literal(text="user@host:~$ ")
    prompt_container.append(prompt_text)
    input_container.append(prompt_container)

    command_container = nodes.inline()
    command_container["classes"] = "command"
    command = nodes.literal(text="echo 'hello'\n")
    command_container.append(command)
    input_container.append(command_container)
    expected.append(input_container)

    output_block = nodes.literal_block(text="hello")
    output_block["classes"] = "terminal-code"
    output_block["xml:space"] = "preserve"
    expected.append(output_block)

    actual = fake_terminal_directive.run()[0]

    assert str(expected) == str(actual)


@pytest.mark.parametrize(
    "fake_terminal_directive",
    [
        {
            "content": ["echo 'hello'", "echo 'test'", "", "hello", "test"],
        }
    ],
    indirect=True,
)
def test_terminal_multiline(fake_terminal_directive):
    expected = nodes.container()
    expected["classes"] = "terminal"

    highlight = addnodes.highlightlang()
    highlight["force"] = "False"
    highlight["lang"] = "text"
    highlight["linenothreshold"] = "10000"
    expected.append(highlight)

    input_container = nodes.container()
    input_container["classes"] = "input"

    prompt_container = nodes.container()
    prompt_container["classes"] = "prompt"
    prompt_text = nodes.literal(text="user@host:~$ ")
    prompt_container.append(prompt_text)
    input_container.append(prompt_container)

    command_container = nodes.inline()
    command = nodes.literal(text="echo 'hello'\n")
    command_container.append(command)
    command = nodes.literal(text="> echo 'test'\n")
    command_container.append(command)
    command_container["classes"] = "command"
    input_container.append(command_container)
    expected.append(input_container)

    output_block = nodes.literal_block(text="hello\ntest")
    output_block["classes"] = "terminal-code"
    output_block["xml:space"] = "preserve"
    expected.append(output_block)

    actual = fake_terminal_directive.run()[0]

    assert str(expected) == str(actual)


@pytest.mark.parametrize(
    "fake_terminal_directive",
    [
        {
            "options": {"output-only": None},
            "content": ["hello"],
        }
    ],
    indirect=True,
)
def test_terminal_no_input(fake_terminal_directive):
    expected = nodes.container()
    expected["classes"] = "terminal"

    highlight = addnodes.highlightlang()
    highlight["force"] = "False"
    highlight["lang"] = "text"
    highlight["linenothreshold"] = "10000"
    expected.append(highlight)

    output_block = nodes.literal_block(text="hello")
    output_block["classes"] = "terminal-code"
    output_block["xml:space"] = "preserve"
    expected.append(output_block)

    actual = fake_terminal_directive.run()[0]

    assert str(expected) == str(actual)


@pytest.mark.parametrize(
    "fake_terminal_directive",
    [
        {
            "content": ["echo 'hello'", "echo 'test'"],
        }
    ],
    indirect=True,
)
def test_terminal_no_output(fake_terminal_directive):
    expected = nodes.container()
    expected["classes"] = "terminal"

    highlight = addnodes.highlightlang()
    highlight["force"] = "False"
    highlight["lang"] = "text"
    highlight["linenothreshold"] = "10000"
    expected.append(highlight)

    input_container = nodes.container()
    input_container["classes"] = "input"

    prompt_container = nodes.container()
    prompt_container["classes"] = "prompt"
    prompt_text = nodes.literal(text="user@host:~$ ")
    prompt_container.append(prompt_text)
    input_container.append(prompt_container)

    command_container = nodes.inline()
    command = nodes.literal(text="echo 'hello'\n")
    command_container.append(command)
    command = nodes.literal(text="> echo 'test'\n")
    command_container.append(command)
    command_container["classes"] = "command"
    input_container.append(command_container)
    expected.append(input_container)

    actual = fake_terminal_directive.run()[0]

    assert str(expected) == str(actual)


@pytest.mark.parametrize(
    "fake_terminal_directive",
    [
        {
            "options": {
                "copy": None,
                "output-only": None,
            },
        }
    ],
    indirect=True,
)
def test_terminal_copy_output_only(fake_terminal_directive):
    with pytest.raises(SphinxError):
        fake_terminal_directive.run()



# This file is part of sphinx-terminal.
#
# Copyright 2025 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License version 3, as published by the Free Software
# Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranties of MERCHANTABILITY, SATISFACTORY
# QUALITY, or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.

"""The core elements of the sphinx-terminal directive."""

from docutils import nodes
from docutils.parsers.rst import directives
from sphinx import addnodes
from sphinx.errors import SphinxError
from sphinx.util.docutils import SphinxDirective


class TerminalDirective(SphinxDirective):
    """Define the terminal directive's state and behavior."""

    required_arguments = 0
    optional_arguments = 0
    has_content = True
    option_spec = {
        "class": directives.class_option,
        "user": directives.unchanged,
        "host": directives.unchanged,
        "dir": directives.unchanged,
        "output-only": directives.flag,
        "scroll": directives.flag,
        "copy": directives.flag,
    }

    @staticmethod
    def input_line(
        prompt_text: str,
        commands: list[str],
        *,
        is_copyable: bool,
    ) -> nodes.container:
        """Construct the prompt with the user-provided options (if any)."""
        input_line = nodes.container()
        input_line["classes"].append("input")

        # To let the prompt be styled separately in LaTeX, it needs to be
        # wrapped in a container. This adds an extra div to the HTML output,
        # but what's a few bytes between friends?
        prompt_container = nodes.container()
        prompt_container["classes"].append("prompt")

        prompt = nodes.literal(text=prompt_text)
        prompt_container.append(prompt)
        input_line.append(prompt_container)

        command = nodes.inline()

        for line in commands:
            command.append(nodes.literal(text=f"{line}\n"))

        command["classes"].append("command")
        if is_copyable:
            command["classes"].append("copybutton")

        input_line.append(command)
        return input_line

    def run(self) -> list[nodes.Node]:
        """Construct the output of the terminal directive."""
        # if :user: or :host: are provided, replace those in the prompt
        classes = self.options.get("class", "")
        user = self.options.get("user", "user")
        host = self.options.get("host", "host")
        prompt_dir = self.options.get("dir", "~")
        user_symbol = "#" if user == "root" else "$"
        has_input = "output-only" not in self.options

        # Raise an error if :copy: and :output-only: are both declared
        if "output-only" in self.options and "copy" in self.options:
            raise SphinxError(
                ":copy: and :output-only: are mutually incompatible. Only input can be copied."
            )

        if user and host:
            prompt_text = f"{user}@{host}:{prompt_dir}{user_symbol} "
        elif user and not host:
            # Only the user is supplied
            prompt_text = f"{user}:{prompt_dir}{user_symbol} "
        else:
            # Omit both user and host, just showing the host
            # doesn't really make sense
            prompt_text = f"{prompt_dir}{user_symbol} "

        out = nodes.container()
        out["classes"].append("terminal")
        for item in classes:
            out["classes"].append(item)
        if "scroll" in self.options:
            out["classes"].append("scroll")

        # The super-large value for linenothreshold is a major hack since I
        # can't figure out how to disable line numbering and the
        # linenothreshold kwarg seems to be required.
        out.append(
            addnodes.highlightlang(lang="text", force=False, linenothreshold=10000)
        )

        # Split inputs and output
        input_lines: list[str] = []
        output_lines: list[str] = []

        # Add the prompt and input
        for line in self.content:
            if has_input:
                if has_input := bool(line.strip()):
                    input_lines += [line]
            else:
                output_lines += [line]

        if input_lines:
            out.append(
                self.input_line(
                    prompt_text,
                    [input_lines[0]] + ["> " + line for line in input_lines[1:]],
                    is_copyable="copy" in self.options,
                )
            )

        # Add output lines
        if output_lines:
            output = nodes.literal_block(text="\n".join(output_lines))
            output["classes"].append("terminal-code")
            out.append(output)

        return [out]
